#include "StdAfx.h"
#include "phys_misc.h"

float CAABB::GetMin( UINT axis ) const
{
	return axis == 0 ? this->Left() : this->Top();
}

float CAABB::GetMax( UINT axis ) const
{
	return axis == 0 ? Right() : Bottom();
}

ASAP_AABB CAABB::GetASAP_AABB()
{
	ASAP_AABB box;
	box.mMax.x = Right();
	box.mMax.y = Bottom();
	box.mMax.z = 1;
	box.mMin.x = Left();
	box.mMin.y = Top();
	box.mMin.z = 0;
	return box;
}

CAABB::CAABB( const CAABB& a )
{
	p = a.p;
	H = a.H;
	W = a.W;
}

CAABB::CAABB( float x1, float y1, float x2, float y2 )
{
	//H = fabs(y2 - y1) * 0.5f;
	//W = fabs(x2 - x1) * 0.5f;
	W = (max(x1,x2) - min(x1, x2)) * 0.5f;
	H = (max(y1,y2) - min(y1, y2)) * 0.5f;
	p = Vector2(min(x1, x2) + W, min(y1, y2) + H);
}

bool CAABB::IsEmpty()
{
	return (H == 0 && W == 0);
}

void CAABB::Intersect( const CAABB &a, const CAABB &b )
{
	if (fabs(a.p.x - b.p.x) <= a.W + b.W &&
		fabs(a.p.y - b.p.y) <= a.H + b.H)
	{
		float al, bl, ar, br, at, bt, ab, bb;
		float cl, cr, cb, ct;

		al = a.p.x - a.W;
		at = a.p.y - a.H;
		bl = b.p.x - b.W;
		bt = b.p.y - b.W;

		ar = a.p.x + a.W;
		ab = a.p.y + a.H;
		br = b.p.x + b.W;
		bb = b.p.y + b.W;

		cl = al > bl ? al : bl;
		cr = ar < br ? ar : br;
		ct = at > bt ? at : bt;
		cb = ab < bb ? ab : bb;

		W = (cl - cr) / 2;
		H = (cb - ct) / 2;
		p.x = cl + W;
		p.y = ct + H;
	}
	else
	{
		H = 0;
		W = 0;
		p = Vector2(0, 0);
	}
}

bool CAABB::PointInCAABB( float x, float y )
{
	return (x >= this->Left() && x <= this->Right() && y >= this->Top() && y <= this->Bottom() );
}

bool CAABB::operator ==( const CAABB &A ) const
{
	return p == A.p && W == A.W && H == A.H;
}

bool CAABB::operator !=( const CAABB &A ) const
{
	return !(*this == A);
}

//////////////////////////////////////////////////////////////////////////
// CRay

CRay::CRay( const CRay& r )
{
	p = r.p;
	k = r.k;
	dir = r.dir;
}

CRay::CRay( float x0, float y0, float k, bool dir )
{
	p = Vector2(x0, y0);
	this->k = k;
	this->dir = dir;

	//      dir.
	assert(!IsVertical() || (k > 0) == dir );
}

CRay::CRay( float x0, float y0, float x1, float y1 )
{
	p = Vector2(x0, y0);
	if (x1 != x0)
	{
		k = (y1-y0)/(x1-x0);
		dir = x1 > x0;
	}
	else
	{
		k = (y1 >= y0) ? 
			numeric_limits<float>::infinity() : 
		-numeric_limits<float>::infinity();
		dir = k > 0;
	}
}

bool CRay::IsVertical()
{
	// TODO: 
	return isinf(k) || isinf(-k);
}

bool CRay::IsHorizontal()
{
	// TODO:   IsNear
	return fabs(k) < 0.00001;
}

bool CRay::IsIntersecting( float left, float right, float top, float bottom )
{
	assert(top < bottom);
	assert(left < right);

	//    .      
	if (p.x < left)
	{
		//   

		if (IsVertical() || !dir)
			return false;

		if ( IsHorizontal() )
			return top <= p.y && p.y <= bottom && dir;

		if (k > 0)
			//   
			return IsHorLineInters(left, right, top) || IsVertLineInters(top, bottom, left);
		else
			//   
			return IsHorLineInters(left, right, bottom) || IsVertLineInters(top, bottom, left);
	}
	else if (p.x > right)
	{
		//   
		if (IsVertical() || dir)
			return false;

		if ( IsHorizontal() )
			return top <= p.y && p.y <= bottom && !dir;

		if (k > 0)
			//   
			return IsHorLineInters(left, right, bottom) || IsVertLineInters(top, bottom, right);
		else
			//   
			return IsHorLineInters(left, right, top) || IsVertLineInters(top, bottom, right);
	}
	else
	{
		//   (   )
		if (p.y < top)
		{
			//  
			if (IsHorizontal() || (IsVertical() && !dir))
				return false;
			else if ((dir && k < 0) || (!dir && k > 0))
				return false;
			else
				return IsHorLineInters(left, right, top);
		}
		else if (p.y > bottom)
		{
			//  
			if (IsHorizontal() || (IsVertical() && dir))
				return false;
			else if ((dir && k > 0) || (!dir && k < 0))
				return false;
			else
				return IsHorLineInters(left, right, bottom);
		}
		else
		{
			//  ,    
			return true;
		}
	}

	return false;
}

bool CRay::IsIntersecting( const CAABB& aabb )
{
	//float left = aabb.Left();
	//float right = aabb.Right();
	//float top = aabb.Top();
	//float bottom = aabb.Bottom();

	return IsIntersecting(aabb.Left(), aabb.Right(), aabb.Top(), aabb.Bottom());
}

bool CRay::GetIntersectionPoint( CAABB const& aabb, Vector2& p )
{
	return GetIntersectionPoint(aabb.Left(), aabb.Top(), aabb.Right(), aabb.Bottom(), p);
}

bool CRay::GetIntersectionPoint( float left, float top, float right, float bottom, Vector2& c )
{
	assert(top < bottom);
	assert(left < right);

	if (p.x < left)
	{
		//   

		if (IsVertical() || !dir)
			return false;

		if ( IsHorizontal() )
			return dir && GetVertLineInters(top, bottom, left, c);

		if (k > 0)
			//   
			return GetHorLineInters(left, right, top, c) || GetVertLineInters(top, bottom, left, c);
		else
			//   
			return GetHorLineInters(left, right, bottom, c) || GetVertLineInters(top, bottom, left, c);
	}
	else if (p.x > right)
	{
		//   
		if (IsVertical() || dir)
			return false;

		if ( IsHorizontal() )
			return !dir && GetVertLineInters(top, bottom, right, c);

		if (k > 0)
			//   
			return GetHorLineInters(left, right, bottom, c) || GetVertLineInters(top, bottom, right, c);
		else
			//   
			return GetHorLineInters(left, right, top, c) || GetVertLineInters(top, bottom, right, c);
	}
	else
	{
		//   (   )
		if (p.y < top)
		{
			//  
			if (IsHorizontal() || (IsVertical() && !dir))
				return false;
			else if ((dir && k < 0) || (!dir && k > 0))
				return false;
			else
				return GetHorLineInters(left, right, top, c);
		}
		else if (p.y > bottom)
		{
			//  
			if (IsHorizontal() || (IsVertical() && dir))
				return false;
			else if ((dir && k > 0) || (!dir && k < 0))
				return false;
			else
				return GetHorLineInters(left, right, bottom, c);
		}
		else
		{
			if (IsVertical())
			{
				if (dir)
					return GetHorLineInters(left, right, top, c);
				else
					return GetHorLineInters(left, right, bottom, c);
			}

			if (dir)
			{
				if (k > 0)
					//   
					return GetHorLineInters(left, right, bottom, c) || GetVertLineInters(top, bottom, right, c);
				else
					//   
					return GetHorLineInters(left, right, top, c) || GetVertLineInters(top, bottom, right, c);
			}
			else
			{
				if (k > 0)
					//   
					return GetHorLineInters(left, right, top, c) || GetVertLineInters(top, bottom, left, c);
				else
					//   
					return GetHorLineInters(left, right, bottom, c) || GetVertLineInters(top, bottom, left, c);
			}
		}
	}

	return false;
}

bool CRay::GetIntersectionPoint( float x1, float y1, float x2, float y2, float& x, float& y )
{
	float a2 = y2 - y1;
	float b2 = x1 - x2;
	float c2 = x2*y1 - x1*y2;

	if (IsVertical())
	{
		x = p.x;
		y = - (a2*x + c2) / b2;

		if (min(x1, x2) - x > 0.01 || x - max(x1,x2) > 0.01 || 
			(y < p.y && k > 0) || (y > p.y && k < 0))
			return false;
	}
	else
	{
		y = ( a2 * (p.y - k*p.x) - k * c2 ) / (b2*k + a2);
		if (min(y1, y2) - y > 0.01 || y - max(y1,y2) > 0.01)
			return false;

		if (k != 0)
			x = (y - p.y) / k + p.x;
		else if (a2 != 0)
			x =  ( b2*y - c2 ) / a2;
		else
		{
			x = dir ? min(x1, x2) : max(x1, x2);
		}

		if (min(x1, x2) - x > 0.01 || x - max(x1,x2) > 0.01 || 
			(x < p.x && dir) || (x > p.x && !dir))
			return false;
	}

	return true;
}

bool CRay::IsHorLineInters( float x1, float x2, float y )
{
	assert(x1 <= x2);
	float xn = (y - p.y) / k + p.x;
	return x1 <= xn && xn <= x2;
}

bool CRay::IsVertLineInters( float y1, float y2, float x )
{
	assert(y1 <= y2);
	float yn = (x - p.x) * k + p.y;
	return y1 <= yn && yn <= y2;
}

bool CRay::GetHorLineInters( float x1, float x2, float y, Vector2& c )
{
	assert(x1 <= x2);
	c.x = (y - p.y) / k + p.x;
	c.y = y;
	return x1 <= c.x && c.x <= x2;
}

bool CRay::GetVertLineInters( float y1, float y2, float x, Vector2& c )
{
	assert(y1 <= y2);
	c.y = (x - p.x) * k + p.y;
	c.x = x;
	return y1 <= c.y && c.y <= y2;
}
